#include "cluster.h"
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include "vector.h"


#define EPSILON 0.001


cluster_int_t* cluster_create_int(vector_int_t* center) {
	cluster_int_t* cluster = malloc(sizeof(cluster_int_t));
	if (NULL == cluster) return NULL;
	cluster->center = center;
	cluster->points = list_points_create_int();
	return cluster;
}

cluster_fpt_t* cluster_create_fpt(vector_fpt_t* center) {
	cluster_fpt_t* cluster = malloc(sizeof(cluster_fpt_t));
	if (NULL == cluster) return NULL;
	cluster->center = center;
	cluster->points = list_points_create_fpt();
	return cluster;
}


void cluster_destroy_int(cluster_int_t* cluster) {
	if (NULL == cluster) return;
	vector_destroy_int(cluster->center);
	list_points_destroy_int(cluster->points, false);
	free(cluster);
}

void cluster_destroy_fpt(cluster_fpt_t* cluster) {
	if (NULL == cluster) return;
	vector_destroy_fpt(cluster->center);
	list_points_destroy_fpt(cluster->points, false);
	free(cluster);
}


void cluster_add_point_int(cluster_int_t* cluster, vector_int_t* point) {
	if (NULL == cluster || NULL == point) return;
	list_points_append_int(cluster->points, point);
}

void cluster_add_point_fpt(cluster_fpt_t* cluster, vector_fpt_t* point) {
	if (NULL == cluster || NULL == point) return;
	list_points_append_fpt(cluster->points, point);
}


bool cluster_update_center_int(cluster_int_t* cluster) {
	// save old center
	vector_int_t* old_center = cluster->center;
	assert(old_center != NULL);
	assert(cluster != NULL);
	// create new center
	list_points_node_int_t* node = cluster->points->head;
	// if cluster is empty
	if (NULL == node) {
		return false;  // center has not been changed
	} else {
		cluster->center = vector_create_int(node->point->dim);
		// sum all values in center
		while (node != NULL) {
			vector_add_inplace_int(cluster->center, *(node->point));
			node = node->next;
		}
		// divide by number of points
		vector_div_inplace_int(cluster->center, (int_t) cluster->points->size);
		// check whether center has changed
		bool changed = false;
		for (size_t p = 0; p < cluster->center->dim; ++p) {
			if (cluster->center->data[p] == old_center->data[p]) {
				changed = true;
				break;
			}
		}
		// destroy old center
		vector_destroy_int(old_center);
		// return true if center has changed
		return changed;
	}
}

bool cluster_update_center_fpt(cluster_fpt_t* cluster) {
	// save old center
	vector_fpt_t* old_center = cluster->center;
	assert(old_center != NULL);
	assert(cluster != NULL);
	// create new center
	list_points_node_fpt_t* node = cluster->points->head;
	// if cluster is empty
	if (NULL == node) {
		return false;  // center has not been changed
	} else {
		cluster->center = vector_create_fpt(node->point->dim);
		// sum all values in center
		while (node != NULL) {
			vector_add_inplace_fpt(cluster->center, *(node->point));
			node = node->next;
		}
		// divide by number of points
		vector_div_inplace_fpt(cluster->center, (fpt_t) cluster->points->size);
		// check whether center has changed
		bool changed = false;
		for (size_t p = 0; p < cluster->center->dim; ++p) {
			if (fabs(cluster->center->data[p] - old_center->data[p]) < EPSILON) {
				changed = true;
				break;
			}
		}
		// destroy old center
		vector_destroy_fpt(old_center);
		// return true if center has changed
		return changed;
	}
}


void cluster_reset_int(cluster_int_t* cluster) {
	list_points_destroy_int(cluster->points, false);
	cluster->points = list_points_create_int();
}

void cluster_reset_fpt(cluster_fpt_t* cluster) {
	list_points_destroy_fpt(cluster->points, false);
	cluster->points = list_points_create_fpt();
}
