#include "kmeans.h"
#include <assert.h>
#include "vector.h"


bool is_vector_in_centers_int(const vector_int_t* center, const cluster_int_t** clusters, const size_t i) {
	for (size_t k = 0; k < i; ++k) {
		if (vector_equals_int(clusters[k]->center, center)) {
			return true;
		}
	}
	return false;
}

cluster_int_t** kmeans_init_clusters_int(const vector_int_t** points, const size_t point_count, const size_t nclusters) {
	// check args and init
	if (NULL == points || point_count < 2 || nclusters < 2) return NULL;
	cluster_int_t** clusters = calloc(nclusters, sizeof(vector_int_t*));
	if (NULL == clusters) return NULL;
	for (size_t k = 0; k < nclusters; ++k) {
		clusters[k] = cluster_create_int(NULL);
	}
	// determine range in which we are working
	vector_int_t* min = vector_copy_int(points[0]);
	vector_int_t* max = vector_copy_int(points[0]);
	for (size_t i = 1; i < point_count; ++i) {
		for (size_t p = 0; p < max->dim; ++p) {
			const int_t value = points[i]->data[p];
			if (value < min->data[p]) min->data[p] = value;
			if (value > max->data[p]) max->data[p] = value;
		}
	}
	// until we have enough centers
	for (size_t i = 0; i < nclusters; ++i) {
		vector_int_t* center = vector_create_int(max->dim);
		bool valid = false;
		while (!valid) {
			// initialise center values randomly, within the "multidimensional rectangle" of our set of points
			for (size_t p = 0; p < center->dim; ++p) {
				center->data[p] = rand_int_range((int) min->data[p], (int) max->data[p]);
			}
			// check center is not already in clusters, although probability is extremely low...
			valid = !is_vector_in_centers_int(center, (const cluster_int_t**) clusters, i);
		}
		cluster_destroy_int(clusters[i]);
		clusters[i] = cluster_create_int(center);
	}
	// free
	vector_destroy_int(min);
	vector_destroy_int(max);
	return clusters;
}

bool is_vector_in_centers_fpt(const vector_fpt_t* center, const cluster_fpt_t** clusters, const size_t i) {
	for (size_t k = 0; k < i; ++k) {
		if (vector_equals_fpt(clusters[k]->center, center)) {
			return true;
		}
	}
	return false;
}

cluster_fpt_t** kmeans_init_clusters_fpt(const vector_fpt_t** points, const size_t point_count, const size_t nclusters) {
	// check args and init
	if (NULL == points || point_count < 2 || nclusters < 2) return NULL;
	cluster_fpt_t** clusters = calloc(nclusters, sizeof(vector_fpt_t*));
	if (NULL == clusters) return NULL;
	for (size_t k = 0; k < nclusters; ++k) {
		clusters[k] = cluster_create_fpt(NULL);
	}
	// determine range in which we are working
	vector_fpt_t* min = vector_copy_fpt(points[0]);
	vector_fpt_t* max = vector_copy_fpt(points[0]);
	for (size_t i = 1; i < point_count; ++i) {
		for (size_t p = 0; p < max->dim; ++p) {
			const fpt_t value = points[i]->data[p];
			if (value < min->data[p]) min->data[p] = value;
			if (value > max->data[p]) max->data[p] = value;
		}
	}
	// until we have enough centers
	for (size_t i = 0; i < nclusters; ++i) {
		vector_fpt_t* center = vector_create_fpt(max->dim);
		bool valid = false;
		while (!valid) {
			// initialise center values randomly, within the "multidimensional rectangle" of our set of points
			for (size_t p = 0; p < center->dim; ++p) {
				center->data[p] = rand_double_range(min->data[p], max->data[p]);
			}
			// check center is not already in clusters, although probability is extremely low...
			valid = !is_vector_in_centers_fpt(center, (const cluster_fpt_t**) clusters, i);
		}
		cluster_destroy_fpt(clusters[i]);
		clusters[i] = cluster_create_fpt(center);
	}
	// free
	vector_destroy_fpt(min);
	vector_destroy_fpt(max);
	return clusters;
}


void kmeans_destroy_clusters_int(cluster_int_t** clusters, const size_t nb_clusters) {
	if (NULL == clusters) return;
	for (size_t i = 0; i < nb_clusters; ++i) {
		cluster_destroy_int(clusters[i]);
	}
}

void kmeans_destroy_clusters_fpt(cluster_fpt_t** clusters, const size_t nb_clusters) {
	if (NULL == clusters) return;
	for (size_t i = 0; i < nb_clusters; ++i) {
		cluster_destroy_fpt(clusters[i]);
	}
}


void kmeans_int(vector_int_t** points, const size_t point_count, cluster_int_t** clusters, const size_t nb_clusters,
				fpt_t (* distance_function)(const vector_int_t*, const vector_int_t*)) {
	bool changed = true;
	while (changed) {
		// reset condition
		changed = false;
		// empty all clusters, keeping only their centers (virtual)
		for (size_t k = 0; k < nb_clusters; ++k) {
			cluster_reset_int(clusters[k]);
		}
		// for each point
		for (size_t i = 0; i < point_count; ++i) {
			vector_int_t* point = points[i];
			// find closest cluster
			cluster_int_t* cmin = clusters[0];
			fpt_t dmin = distance_function(point, cmin->center);
			for (size_t k = 0; k < nb_clusters; ++k) {
				cluster_int_t* current_cluster = clusters[k];
				fpt_t dist = distance_function(point, current_cluster->center);
				if (dist < dmin) {
					cmin = current_cluster;
					dmin = dist;
				}
			}
			// add point to closest cluster
			cluster_add_point_int(cmin, point);
		}
		// update all cluster centers
		for (size_t k = 0; k < nb_clusters; ++k) {
			assert(clusters[k] != NULL);
			assert(clusters[k]->points != NULL);
			if (cluster_update_center_int(clusters[k])) {
				changed = true;
			}
		}
	}
}


void kmeans_fpt(vector_fpt_t** points, const size_t point_count, cluster_fpt_t** clusters, const size_t nb_clusters,
				fpt_t (* distance_function)(const vector_fpt_t*, const vector_fpt_t*)) {
	bool changed = true;
	while (changed) {
		// reset condition
		changed = false;
		// empty all clusters, keeping only their centers (virtual)
		for (size_t k = 0; k < nb_clusters; ++k) {
			cluster_reset_fpt(clusters[k]);
		}
		// for each point
		for (size_t i = 0; i < point_count; ++i) {
			vector_fpt_t* point = points[i];
			// find closest cluster
			cluster_fpt_t* cmin = clusters[0];
			fpt_t dmin = distance_function(point, cmin->center);
			for (size_t k = 0; k < nb_clusters; ++k) {
				cluster_fpt_t* current_cluster = clusters[k];
				fpt_t dist = distance_function(point, current_cluster->center);
				if (dist < dmin) {
					cmin = current_cluster;
					dmin = dist;
				}
			}
			// add point to closest cluster
			cluster_add_point_fpt(cmin, point);
		}
		// update all cluster centers
		for (size_t k = 0; k < nb_clusters; ++k) {
			assert(clusters[k] != NULL);
			assert(clusters[k]->points != NULL);
			if (cluster_update_center_fpt(clusters[k])) {
				changed = true;
			}
		}
	}
}
