#include <assert.h>
#include <getopt.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "distance.h"
#include "io.h"
#include "kmeans.h"
#include "linkedlist.h"
#include "vector.h"


enum DistanceFunctionType {
	EUCLID = 0, MANHATTAN = 1, CHEBYSHEV = 2
};

enum DataType {
	FLOAT = 0, INT = 1
};


fpt_t (* const DIST_FUNC_INT[])(const vector_int_t*, const vector_int_t*) = {
		distance_euclid_int,
		distance_manhattan_int,
		distance_chebyshev_int};

fpt_t (* const DIST_FUNC_FPT[])(const vector_fpt_t*, const vector_fpt_t*) = {
		distance_euclid_fpt,
		distance_manhattan_fpt,
		distance_chebyshev_fpt};


void help(const char* callname) {
	fprintf(stderr,
			"\nUSAGE:    %s  -i INPUT_FILE  -o OUTPUT_FILE  -d [euclid,manhattan,chebyshev]  -t [fpt,int]\n",
			callname);
}


void parse_args(int argc, char** argv, char** ipath, char** opath, enum DistanceFunctionType* df, enum DataType* type) {
	int opt;
	while ((opt = getopt(argc, argv, "i:o:d:t:h")) != -1) {
		switch (opt) {
			case 'h':
				help(argv[0]);
				exit(EXIT_FAILURE);
			case 'i':
				*ipath = optarg;
				break;
			case 'o':
				*opath = optarg;
				break;
			case 'd':
				if (strcmp(optarg, "euclid") == 0) *df = EUCLID;
				else if (strcmp(optarg, "manhattan") == 0) *df = MANHATTAN;
				else if (strcmp(optarg, "chebyshev") == 0) *df = CHEBYSHEV;
				break;
			case 't':
				if (strcmp(optarg, "fpt") == 0) *type = FLOAT;
				else if (strcmp(optarg, "int") == 0) *type = INT;
				break;
			case '?':
				fprintf(stderr, "UNKNOWN OPTION :  %c", opt);
				break;
			default:
				// https://www.gnu.org/software/libc/manual/html_node/Example-of-Getopt.html
				abort();
		}
	}
}


int main_int(FILE* ifile, FILE* ofile, const size_t dim, const size_t nb_clusters, const enum DistanceFunctionType dist_func_type) {
	// INIT
	vector_int_t** points = NULL;
	cluster_int_t** clusters = NULL;
	// READ
	list_points_int_t* list = io_get_vector_list_int(ifile, dim);
	const size_t point_count = list->size;
	points = list_points_to_array_int(list);
	list_points_destroy_int(list, false);
	list = NULL;
	// ALGORITHM
	printf("INIT:   ");
	clusters = kmeans_init_clusters_int((const vector_int_t**) points, point_count, nb_clusters);
	printf("DONE\n");
	printf("KMEANS: begin\n");
	kmeans_int(points, point_count, clusters, nb_clusters, DIST_FUNC_INT[dist_func_type]);
	printf("KMEANS: DONE !\n");
	// WRITE
	fprintf(ofile, "%lu\n%lu\n", dim, nb_clusters);
	io_write_clusters_to_file_int(ofile, clusters, nb_clusters);
	// CLEANUP
	for (size_t i = 0; i < nb_clusters; ++i) cluster_destroy_int(clusters[i]);
	for (size_t i = 0; i < point_count; ++i) vector_destroy_int(points[i]);
	// EXIT
	return EXIT_SUCCESS;
}

int main_fpt(FILE* ifile, FILE* ofile, const size_t dim, const size_t nb_clusters, const enum DistanceFunctionType dist_func_type) {
	// INIT
	vector_fpt_t** points = NULL;
	cluster_fpt_t** clusters = NULL;
	// READ
	list_points_fpt_t* list = io_get_vector_list_fpt(ifile, dim);
	const size_t point_count = list->size;
	points = list_points_to_array_fpt(list);
	list_points_destroy_fpt(list, false);
	list = NULL;
	// ALGORITHM
	printf("INIT:   ");
	clusters = kmeans_init_clusters_fpt((const vector_fpt_t**) points, point_count, nb_clusters);
	printf("DONE\n");
	printf("KMEANS: begin\n");
	kmeans_fpt(points, point_count, clusters, nb_clusters, DIST_FUNC_FPT[dist_func_type]);
	printf("KMEANS: DONE !\n");
	// WRITE
	fprintf(ofile, "%lu\n%lu\n", dim, nb_clusters);
	io_write_clusters_to_file_fpt(ofile, clusters, nb_clusters);
	// CLEANUP
	for (size_t i = 0; i < nb_clusters; ++i) cluster_destroy_fpt(clusters[i]);
	free(clusters);
	for (size_t i = 0; i < point_count; ++i) vector_destroy_fpt(points[i]);
	free(points);
	// EXIT
	return EXIT_SUCCESS;
}


int main(int argc, char** argv) {
	// init defaults
	char (* ipath) = NULL, (* opath) = NULL;
	enum DistanceFunctionType disttype = EUCLID;
	enum DataType datatype = FLOAT;
	// parse args
	parse_args(argc, argv, &ipath, &opath, &disttype, &datatype);
	// open files
	FILE* ifile = ipath != NULL ? fopen(ipath, "r") : stdin;
	FILE* ofile = opath != NULL ? fopen(opath, "w") : stdout;
	// read dimension and desired number of clusters from file
	const size_t dim = io_read_int(ifile);
	const size_t nb_clusters = io_read_int(ifile);
	if (0 == dim) {
		printf("DIMENSION MUST BE STRICTLY POSITIVE !\n");
		fclose(ifile);
		fclose(ofile);
		return EXIT_FAILURE;
	}
	if (0 == nb_clusters) {
		printf("NUMBER OF CLUSTERS MUST BE STRICTLY POSITIVE !\n");
		fclose(ifile);
		fclose(ofile);
		return EXIT_FAILURE;
	}
	// type specific code
	int return_value = EXIT_FAILURE;
	int (* main_routine)(FILE*, FILE*, const size_t, const size_t, const enum DistanceFunctionType);
	main_routine = INT == datatype ? main_int : main_fpt;
	printf(INT == datatype ? "TYPE:   INT\n" : "TYPE:   FLOAT\n");
	assert(ifile != NULL);
	assert(ofile != NULL);
	return_value = main_routine(ifile, ofile, dim, nb_clusters, disttype);
	// cleanup
	fclose(ifile);
	fclose(ofile);
	return return_value;
}
